import asyncio
import ctypes
import re
import inspect

user32 = ctypes.windll.user32

from wizsdk.constants import keycode_map
from wizsdk.client import unregister_all


class HotkeyEvents:
    """
    Hotkey event manager class
    
    Examples:
        .. code-block:: py
        
            import wizsdk
            import asyncio

            # Initiate the event manager
            events = wizsdk.HotkeyEvents()

            # Add hotkeys
            def print_hello():
                print("Hello, world!")

            events.set_hotkey("ctrl + q", events.safe_quit) # safetly quit the program
            events.set_hotkey("SPACEBAR", print_hello)
            events.set_hotkey("Left mouse", lambda: print("click"))


            async def main_function():
                # main script goes here

            # run the two coroutines at the same time
            wizsdk.run_threads(events.listen(), main_function())

    
    """

    def __init__(self, debug=False):
        self._actions = {}
        self._pressed = {}
        self.debug = debug

    def _code_from_str(self, key):
        try:
            return keycode_map[key]
        except KeyError:
            # If this throws an error, it won't be catched
            # This is expected behavior
            return keycode_map[key.upper()]

    def _str_to_keycodes(self, trigger):
        # Split
        keys = re.split("\s*\+\s*", trigger)
        # Conver to keycodes
        return tuple([self._code_from_str(k) for k in keys])

    def _trigger_to_str(self, trigger):
        rev = {a: b for (b, a) in keycode_map.items()}
        return " + ".join([rev[k] for k in trigger])

    def set_hotkey(self, trigger: str, action):
        """
        Registers a hotkey
        
        Args:
            trigger: the hotkey(s) that will trigger the action. Separate multiple keys with a ``+``
            action: the function that will run when the hotkey is triggered. This can be a regular or an ``await``able function
        
        """
        if type(trigger) != str:
            raise ValueError(
                f"Invalid trigger of type {type(trigger)}. Expecting type `str`"
            )

        if not callable(action):
            raise ValueError("Invalid param `action`. Param not callable")

        try:

            keys_as_codes = self._str_to_keycodes(trigger)

            self.debug and print(trigger, keys_as_codes)

            self._actions[keys_as_codes] = action
            # Start as True so that it's not executed on start
            self._pressed[keys_as_codes] = True
        except KeyError:
            print("One or more of the hot keys are invalid:", trigger)
            return False

    def unset_hotkey(self, trigger):
        """
        Removes a previously set hotkey
        
        Args:
            trigger: the same trigger used to register the hotkey
        """
        self._actions.pop(self._str_to_keycodes(trigger), None)
        print(self._actions.keys())

    async def listen(self):
        """
        starts an event loop that will listen for key presses and call actions triggered by the hotkeys.
        """
        while True:
            await asyncio.sleep(0.02)
            for (trigger, action) in self._actions.items():
                all_keys_pressed = all(
                    [user32.GetAsyncKeyState(keycode) for keycode in trigger]
                )
                was_pressed = self._pressed[trigger]
                if not was_pressed and all_keys_pressed:
                    self.debug and print(
                        f"hotkey {self._trigger_to_str(trigger)} triggered"
                    )
                    # Check if the action is a coroutine and needs to be awaited
                    if inspect.iscoroutinefunction(action):
                        await action()
                    else:
                        action()
                    # prevents double events by waiting for the keys to be released before being triggered again
                    self._pressed[trigger] = True

                elif not all_keys_pressed and was_pressed:
                    self._pressed[trigger] = False

    async def safe_quit(self):
        """
        Safetly quit the program by first un-hooking all clients.
        """
        await unregister_all()
        quit()


if __name__ == "__main__":
    from utils import run_threads

    # A few tests
    events = HotkeyEvents(debug=True)
    events.set_hotkey("alt + q", events.safe_quit)
    events.set_hotkey("Left mouse", lambda: print("click"))

    async def main_function():
        while True:
            await asyncio.sleep(1)
            print("waiting")

    run_threads(events.listen(), main_function())



# Third-party imports
import asyncio

# Custom imports
from wizsdk.mouse import Mouse


mouse = Mouse()
Card = None


class Card:
    def __init__(self, client, name, spell_x):
        self.client = client
        self.name = name
        self.spell_x = spell_x
        self.spell_y = 325

    def __str__(self):
        return f"{self.name} at ({self.spell_x}, {self.spell_y})"

    async def enchant(self, spell: Card) -> Card:
        """
        Enchants `spell` with self.
        
        Returns:
            A new Card object representing the new enchanted spell
        
        """
        self.client.log(f"Enchanting {spell.name} with {self.name}")

        card_width = 52
        enchant_is_before = self.spell_x < spell.spell_x
        # click self
        await self.client.mouse.click(
            self.spell_x, self.spell_y, duration=0.3, delay=0.6
        )
        # click spell
        await self.client.mouse.click(
            spell.spell_x, spell.spell_y, duration=0.3, delay=0.6
        )
        # calculate new spell_x of enchanted spell

        new_pos = spell.spell_x + (card_width / 2)
        if enchant_is_before:
            new_pos -= card_width

        new_name = f"{self.name}-{spell.name}"

        await asyncio.sleep(0.2)
        return Card(self.client, new_name, new_pos)

    async def cast(self, target=None):
        """
        Selects spell.
        If target is specified, it clicks on target (0-3 for enemies, 4-7 for allies)
        
        .. code-block:: py
        
            0 1 2 3
            -------
            7 6 5 4
        
        Args:
            target (int, optional): The target to select after clicking the spell
        """
        self.client.log(f"Casting {self.name}")
        await self.client.mouse.click(
            self.spell_x, self.spell_y, duration=0.3, delay=0.6
        )

        if target != None:
            if target < 4:
                x = (174 * target) + 130
                y = 50
            elif target < 8:
                x = (174 * (7 - target)) + 160
                y = 590
            else:
                print(
                    f"Invalid value for target, expect int between 0 - 7, got {target}"
                )
                return False
            await self.client.mouse.click(x, y, duration=0.3, delay=0.6)
            await asyncio.sleep(
                0.5
            )  # Helps when changing the active state of overlapping windows.




# Native imports
import ctypes
from ctypes.wintypes import POINT
import time
import asyncio

# Custom imports
from wizsdk.window import Window, screen_size

# If the mouse is over a coordinate in FAILSAFE_POINTS and FAILSAFE is True, the FailSafeException is raised.
# The rest of the points are added to the FAILSAFE_POINTS list at the bottom of this file, after size() has been defined.
# The points are for the corners of the screen, but note that these points don't automatically change if the screen resolution changes.
FAILSAFE = True
FAILSAFE_POINTS = [(0, 0)]
MINIMUM_DURATION = 0.1
MINIMUM_SLEEP = 0.02


class FailSafeException(Exception):
    """
    Exception raised when the mouse is over (0, 0). Protactive measure to regain control of your mouse.
    """

    def __init__(
        self,
        message="Mouse fail-safe triggered from mouse moving to a corner of the screen.",
    ):
        self.message = message
        super().__init__(self.message)


def getPointOnLine(x1, y1, x2, y2, n):
    """
    Returns an (x, y) tuple of the point that has progressed a proportion ``n`` along the line defined by the two
    ``x1``, ``y1`` and ``x2``, ``y2`` coordinates.
    This function was copied from pytweening module, so that it can be called even if PyTweening is not installed.
    """
    x = ((x2 - x1) * n) + x1
    y = ((y2 - y1) * n) + y1
    return (x, y)


class Mouse(Window):
    """
    Class for controlling the computer's mouse.
    """

    _MOUSEEVENTF_MOVE = 0x0001  # mouse move
    _MOUSEEVENTF_LEFTDOWN = 0x0002  # left button down
    _MOUSEEVENTF_LEFTUP = 0x0004  # left button up
    _MOUSEEVENTF_RIGHTDOWN = 0x0008  # right button down
    _MOUSEEVENTF_RIGHTUP = 0x0010  # right button up
    _MOUSEEVENTF_MIDDLEDOWN = 0x0020  # middle button down
    _MOUSEEVENTF_MIDDLEUP = 0x0040  # middle button up
    _MOUSEEVENTF_WHEEL = 0x0800  # wheel button rolled
    _MOUSEEVENTF_ABSOLUTE = 0x8000  # absolute move
    _SM_CXSCREEN = 0
    _SM_CYSCREEN = 1

    def __init__(self, window_handle=None, silent_mode=False, walker=None):
        """
        Args:
            silent_mode: When enabled, moves the mouse without taking control of the actual cursor
            walker: A wizwalker client. It is required for silent mode
        """
        super().__init__(window_handle)
        # Window handle to which the mouse events will be relative to
        self.window_handle = window_handle
        self.walker = walker

        if silent_mode and walker == None:
            raise ValueError("A walker must be passed to Mouse to use silent mode")

        self.silent_mode = silent_mode
        self.silent_xpos = 0
        self.silent_ypos = 0
        self.silent_init = False

    def _do_event(self, flags, x_pos, y_pos, data, extra_info):
        """generate a mouse event"""
        if not self.silent_mode:
            x_calc = int(
                65536 * x_pos / ctypes.windll.user32.GetSystemMetrics(self._SM_CXSCREEN)
                + 1
            )
            y_calc = int(
                65536 * y_pos / ctypes.windll.user32.GetSystemMetrics(self._SM_CYSCREEN)
                + 1
            )
            return ctypes.windll.user32.mouse_event(
                flags, x_calc, y_calc, data, extra_info
            )
        else:
            raise RuntimeError(
                "Mouse._do_event is not supported with silent mode enabled"
            )

    def _get_button_value(self, button_name, button_up=False):
        """convert the name of the button into the corresponding value"""
        buttons = 0
        if button_name.find("right") >= 0:
            buttons = self._MOUSEEVENTF_RIGHTDOWN
        if button_name.find("left") >= 0:
            buttons += self._MOUSEEVENTF_LEFTDOWN
        if button_name.find("middle") >= 0:
            buttons += self._MOUSEEVENTF_MIDDLEDOWN
        if button_up:
            buttons = buttons << 1
        return buttons

    async def init_silent_mode(self):
        if self.silent_mode and not self.silent_init:
            point = POINT()
            point.x = 100
            point.y = 100
            ctypes.windll.user32.ClientToScreen(self.window_handle, ctypes.byref(point))
            self.silent_xpos = point.x
            self.silent_ypos = point.y
            await self.walker.set_mouse_position(
                self.silent_xpos, self.silent_ypos, convert_from_client=False
            )
            self.silent_init = True

    async def _set_position(self, pos):
        """
        Set the position of the mouse to the specified coordinates
        """
        (x, y) = pos

        if not self.silent_mode:
            self._do_event(
                self._MOUSEEVENTF_MOVE + self._MOUSEEVENTF_ABSOLUTE, x, y, 0, 0
            )
        else:
            await self.init_silent_mode()
            self.silent_xpos = x
            self.silent_ypos = y
            await self.walker.set_mouse_position(
                self.silent_xpos, self.silent_ypos, convert_from_client=False
            )

    async def move_to(self, x, y, duration=0.5):
        """
        Move the mouse to the x, y coordinates relative to the window
        """
        # We need to get from (startx, starty) to (x, y)

        # Set the X, Y to be relative to the window position
        await self.init_silent_mode()
        wX, wY, *_ = self.get_rect()

        x += wX
        y += wY

        startx, starty = self.get_position()
        x_offset = x - startx
        y_offset = y - starty

        width, height = screen_size()

        steps = [(x, y)]

        if duration > MINIMUM_DURATION:
            num_steps = max(width, height)
            sleep_amount = duration / num_steps

            if sleep_amount < MINIMUM_SLEEP:
                num_steps = int(duration / MINIMUM_SLEEP)
                sleep_amount = duration / num_steps

            steps = [
                getPointOnLine(startx, starty, x, y, (n / num_steps))
                for n in range(num_steps)
            ]
            steps.append((x, y))

        for _x, _y in steps:
            if len(steps) > 1:
                # A single step doesn't require tweening
                time.sleep(sleep_amount)

            _x = int(round(_x))
            _y = int(round(_y))

            # Failsafe check
            if not self.silent_mode and (_x, _y) not in FAILSAFE_POINTS:
                self.failSafeCheck()

            await self._set_position((_x, _y))

        # Failsafe check
        if not self.silent_mode and not (_x, _y) not in FAILSAFE_POINTS:
            self.failSafeCheck()

    def wizsdk_client_coords_to_wizwalker(self, x: int, y: int) -> (int, int):
        """
        Converts WizSDK client coords to wizwalker client coords
        """
        # make absolute
        wX, wY, *_ = self.get_rect()
        x += wX
        y += wY

        # convert to the client coords wizwalker uses
        point = POINT(int(x), int(y))
        if (
            ctypes.windll.user32.ScreenToClient(self.window_handle, ctypes.byref(point))
            == 0
        ):
            raise RuntimeError("Screen to client conversion failed")

        return (point.x, point.y)

    async def click(self, x=-1, y=-1, button="left", duration=None, delay=0.1):
        """Click at the specified placed"""
        # Set default duration
        if duration is None:
            if x == -1 and y == -1:
                duration = 0
            else:
                duration = 0.5

        # If position is not set, use current mouse position
        old_pos = self.get_position()
        x = x if (x != -1) else old_pos[0]
        y = y if (y != -1) else old_pos[1]
        await self.move_to(x, y, duration=duration)
        await asyncio.sleep(delay)
        if not self.silent_mode:
            self.set_active()
            self._do_event(
                self._get_button_value(button, False)
                + self._get_button_value(button, True),
                0,
                0,
                0,
                0,
            )
        else:
            (nx, ny) = self.wizsdk_client_coords_to_wizwalker(
                x, y
            )  # this is needed because walker.click implicitly converts client to screen, but with a different method
            await self.walker.click(
                nx, ny, right_click=button != "left", sleep_duration=delay
            )

    def double_click(self, pos=(-1, -1), button="left"):
        """Double click at the specifed placed"""
        for i in range(2):
            self.click(pos, button)

    def get_position(self):
        """get mouse position"""
        if not self.silent_mode:
            point = POINT()
            ctypes.windll.user32.GetCursorPos(ctypes.byref(point))
            return (point.x, point.y)
        else:
            return (self.silent_xpos, self.silent_ypos)

    def get_rel_position(self):
        """get mouse position relative to window"""
        if not self.silent_mode:
            wx, wy = self.get_rect()[:2]
            x, y = self.get_position()
            return (x - wx, y - wy)
        else:
            wx, wy = self.get_rect()[:2]
            return (self.silent_xpos - wx, self.silent_ypos - wy)

    def in_rect(self, rect_area):
        """
        Returns true if the mouse is in the given region
        """
        x, y, w, h = rect_area
        wx, wy = self.get_rect()[:2]
        # Make the position of the rect relative to the window
        x += wx
        y += wy

        mouseX, mouseY = self.get_position()
        return mouseX > x and mouseX < (x + w) and mouseY > y and mouseY < (y + h)

    async def move_out(self, rect_area):
        """
        Move the mouse outside of the given rect
        If the mouse is already outside, return
        """
        x, y = self.get_rel_position()
        while self.in_rect(rect_area):
            # TODO find fastest way out of rect
            y -= 100
            await self.move_to(x, y, duration=0.1)

    def failSafeCheck(self):
        if not self.silent_mode and FAILSAFE and self.get_position() in FAILSAFE_POINTS:
            raise FailSafeException


if __name__ == "__main__":
    import asyncio

    async def main():
        mouse = Mouse()
        mouse._set_position((100, 100))
        await mouse.move_to(10, 100, duration=1)

    asyncio.run(main())



# Native imports
import ctypes
import os, sys
import asyncio
from typing import Optional

# Third-party imports
import cv2
import numpy
import wizwalker
from wizwalker.utils import calculate_perfect_yaw

# Custom imports
from .utils import get_all_wiz_handles, XYZYaw, packaged_img
from .pixels import DeviceContext, match_image
from .keyboard import Keyboard
from .mouse import Mouse
from .window import Window
from .battle import Battle
from .card import Card

# rectangles defined as (x, y, width, height)
AREA_FRIENDS = (625, 65, 20, 240)
AREA_SPELLS = (245, 290, 370, 70)
AREA_CONFIRM = (355, 370, 100, 70)

SPELLS_FOLDER = "spells"
""" Default folder to look for spells in"""

IMAGE_FOLDER = ""
""" Default folder to look for images in."""

DEFAULT_MOUNT_SPEED = 1.4
""" Default mount speed (40%) """

user32 = ctypes.windll.user32

# Keep track of all clients
all_clients = []


async def unregister_all():
    """
    Properly unregisters all clients
    """
    print("Un-hooking all registered clients")
    for client in all_clients:
        await client.unregister()


class Client(DeviceContext, Keyboard, Window):
    """
    Main class for wizSDK.

    Example:
        .. code-block:: py

            # registers a new client
            player = client.register(name="My Bot")
            await player.activate_hooks()


    """

    def __init__(self, handle=None, silent_mouse=False):
        # Inherit all methods from Parent Classes
        super().__init__(handle)
        self.window_handle = handle
        self.logging = True
        self.name = None

        self._default_image_folder = IMAGE_FOLDER
        self.walker = None
        self.silent_mouse = silent_mouse
        self.mouse = None

        self._last_health = 99_999
        self._last_mana = 99_999

    @classmethod
    def register(cls, nth=0, name=None, handle=None, silent_mouse: bool = False):
        """
        Assigns the instance to a wizard101 window. (Required before using any other SDK methods)

        Args:
            nth (int, optional): Index of the wizard101 client to use (if there's more than one)
            name (str, optional): Name to prepend to the window title. Used for identifying which windows are controlled.
            silent_mouse: When enabled, moves the mouse without taking control of the actual cursor
        """
        client = cls()
        global all_clients

        client.silent_mouse = silent_mouse
        if handle:
            client.window_handle = handle
        else:
            handles = get_all_wiz_handles()

            if len(handles) == 0:
                client.log("No Wizard101 windows detected")
                return None

            handles.sort()
            # Assigns the one at index nth
            client.window_handle = handles[nth]

        # A window exists, add it to global variable
        all_clients.append(client)

        # Start `_anti_disconnect` task
        client._anti_disconnect_task = asyncio.create_task(client._anti_disconnect())

        client.walker = wizwalker.Client(client.window_handle)

        client.mouse = Mouse(client.window_handle, client.silent_mouse, client.walker)

        if name:
            client.set_name(name)

        return client

    async def activate_hooks(self, *hook_names):
        """
        Activate a number of hooks or pass None/no args to activate all (excluding special hooks like "mouseless_cursor_move")

        Args:
            hook_names: The hooks to activate

        Examples:
            .. code-block:: py

                # activates player_struct and player_stat_struct
                await activate_hooks("player_struct", "player_stat_struct")

                # activates all hooks (excluding special hooks like "mouseless_cursor_move")
                await activate_hooks()
        """
        hooks = hook_names
        if len(hook_names) == 0:
            hooks = tuple(
                [h for h in self.walker.get_hooks() if h != "mouseless_cursor_move"]
            )

        await self.walker.activate_hooks(*hooks)
        await self.send_key("d")
        await self.send_key("a")

    async def activate_all_hooks(self):
        """
        Activate all hooks (including special hooks like "mouseless_cursor_move")
        """
        await self.walker.activate_hooks()
        await self.send_key("d")
        await self.send_key("a")

    def set_name(self, name: str):
        """
        Sets the window title. Useful to identify which window the bot is running on
        Args:
            name (str): The text to prepend to the window title
        """
        self.name = name
        user32.SetWindowTextW(self.window_handle, f"[{name}] Wizard101")

    def log(self, message: str):
        """
        Debug log function. Useful to identify which client the log is coming from.
        Args:
            message (str): The message to log. ``[client-name]: message``
        """
        if self.logging:
            s = ""
            if self.name != None:
                s += f"[{self.name}] "
            s += message
            print(s)
            sys.stdout.flush()

    async def _anti_disconnect(self):
        while True:
            # Wait 10 minute
            await asyncio.sleep(10 * 60)
            # Sent o key to stay awake
            await self.send_key("O", 0.1)
            await self.send_key("O", 0.1)

    async def unregister(self):
        """
        Properly unregister hooks and clean up possible ongoing asyncio tasks
        """
        user32.SetWindowTextW(self.window_handle, "Wizard101")
        self._anti_disconnect_task.cancel()
        await self.walker.close()
        return 1

    async def wait(self, seconds: float):
        """
        Alias for asyncio.sleep()

        Args:
            seconds (float): number of seconds to "sleep"
        """
        await asyncio.sleep(seconds)
        return self

    """
    STATE DETECTION
    """

    async def get_health(self) -> int:
        """
        Gets health value from memory. For accurate values, only use after finishing a fight or after getting whisps. Returns 99,999 if the stats hook hasn't run.

        Returns:
            The health value of the player as an int
        """

        mem_health = await self.walker.health()
        if (mem_health != None) and (mem_health >= 0) and (mem_health < 20_000):
            self._last_health = mem_health

        return self._last_health

    async def get_mana(self) -> int:
        """
        Gets mana value from memory. For accurate values, only use after finishing a fight or after getting whisps. Returns 99,999 if the stats hook hasn't run.

        Returns:
            The mana value of the player as an int
        """

        mem_mana = await self.walker.mana()
        if (mem_mana != None) and (mem_mana >= 0) and (mem_mana < 20_000):
            self._last_mana = mem_mana

        return self._last_mana

    def is_crown_shop(self) -> bool:
        """
        Detects if the crown shop is open by matching a red pixel in the "close" icon.

        Returns:
            bool: True if the menu is open / False otherwise
        """
        return self.pixel_matches_color((788, 53), (197, 40, 41), 50)

    def is_idle(self):
        """
        Detects if the player is idle (out of loading and not in battle) by matching pixels in the spellbook.

        Returns:
            bool: True if the player is idle / False otherwise
        """
        spellbook_yellow = self.pixel_matches_color(
            (781, 551), (255, 251, 64), tolerance=30
        )
        spellbook_brown = self.pixel_matches_color(
            (728, 587), (79, 29, 29), tolerance=30
        )

        spellbook_gray = self.pixel_matches_color(
            (747, 538), (27, 47, 63), tolerance=30
        )
        return spellbook_brown and spellbook_yellow and spellbook_gray

    def is_dialog_more(self):
        """
        Detects if the dialog (from NPCs etc.) is open by matching pixels in the "more" or "done" button.

        Returns:
            bool: True if the dialog menu is open / False otherwise
        """
        more_lower_right = self.pixel_matches_color((669, 611), (112, 32, 54), 15)
        more_top_left = self.pixel_matches_color((585, 609), (115, 32, 56), 15)
        return more_lower_right and more_top_left

    def is_health_low(self):
        """
        DEPRECATED:
        Detects if the player's health is low by matching a red pixel in the lower third of the globe.

        Returns:
            bool: is the health low
        """
        # Matches a pixel in the lower third of the health globe
        POSITION = (23, 563)
        COLOR = (126, 41, 3)
        TOLERANCE = 15
        return not self.pixel_matches_color(POSITION, COLOR, tolerance=TOLERANCE)

    def is_mana_low(self):
        """
        DEPRECATED:
        Detects if the player's mana is low by matching a blue pixel in the lower third of the globe.

        Returns:
            bool: is the mana low
        """
        # Matches a pixel in the lower third of the mana globe
        POSITION = (79, 591)
        COLOR = (66, 13, 83)
        TOLERANCE = 15
        return not self.pixel_matches_color(POSITION, COLOR, tolerance=TOLERANCE)

    def is_press_x(self):
        """
        Detects if the "press x" prompt is present by matching the "x" image.

        Returns:
            bool: "press X" has been found
        """
        x_area = self.get_image(region=(350, 540, 100, 20))
        found = match_image(x_area, packaged_img("x.png"))
        return found != False

    def get_confirm(self):
        """
        Detects if a "confirm" prompt is open (either from teleporting a friend or exiting an unfinished dungeon). It does this by matching a "confirm" image.

        Returns:
            tuple: (x, y) where the confirm button has been found
        """
        confirm_img = self.get_image(AREA_CONFIRM)
        found = match_image(confirm_img, packaged_img("confirm.png"), threshold=0.2)
        if found:
            x = found[0] + AREA_CONFIRM[0]
            y = found[1] + AREA_CONFIRM[1]
            found = (x, y)

        return found

    async def get_backpack_space_left(self) -> Optional[int]:
        """
        Gets the backpack space left. Will try to refresh the value by quickly opening and closing the backpack if necessary. Returns None if it wasn't able to get the value.

        Returns:
            space left in the backpack, None if it's not able to get the value.
        """

        refresh_triggered = False
        time_awaited = 0
        interval = 0.1
        space_used = await self.walker.backpack_space_used()
        while space_used == None and time_awaited < 1 and self.is_idle():
            if not refresh_triggered:
                self.log("Refreshing backpack space value")
                # Open and close character page to force memory update
                await self.send_key("b")
                await self.send_key("b")
                refresh_triggered = True

            # Wait
            await self.wait(interval)
            time_awaited += interval

            # re-try accessing the values
            space_used = await self.walker.backpack_space_used()

        if space_used == None:
            return None

        space_total = await self.walker.backpack_space_total()
        return space_total - space_used

    """
    ACTIONS BASED ON STATES
    """

    async def use_potion_if_needed(self, health=1000, mana=20):
        """
        Clicks on a potion if health or mana values drop below the settings

        Args:
            health: Health value threshold. Potion will be used if health value drops below
            mana:   Mana value threshold. Potion will be used if mana value drops below

        Returns:
            None
        """
        h = await self.get_health()
        m = await self.get_mana()

        mana_low = m < mana
        health_low = h < health

        if mana_low:
            self.log(f"Mana is at {m}, using potion")
        if health_low:
            self.log(f"Health is at {h}, using potion")
        if mana_low or health_low:
            await self.mouse.click(160, 590, delay=0.2)

    async def finish_loading(self, *, timeout=None) -> bool:
        """
        Waits for player to have gone through the loading screen.
        If this function is called too late and the player is already out of the loading screen, it will wait indefinitely.

        Args:
            timeout (optional): value in seconds to timeout if it hasn't finished yet. Defaults to None

        Returns:
            True if the function completed successfully, False if the function timed out.
        """

        async def _loading_coro():
            self.log("Awaiting loading")

            base_addr = lambda: self.walker._memory.process.read_int(
                self.walker._memory.player_struct_addr
            )
            player_struct = base_addr()

            # Wait for base_addr to change
            while player_struct == base_addr():
                await asyncio.sleep(0.2)

            await asyncio.sleep(2)

        # run it with the timeout
        try:
            await asyncio.wait_for(_loading_coro(), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False

    async def go_through_dialog(self, times=1, *, timeout=None):
        # Wait for press X, or more/done button
        """
        Goes through the prompts of the dialog ("press x" or "more"/"continue"). Waits for "press x" or the dialog box before starting.

        Args:
            times (int): Defaults to 1
                The number of times to repeat. (When going through quests, you might need to set this to 2. There's one to hand in the quest, the second to get the next quest)
            timeout (optional): value in seconds to timeout if it hasn't finished yet. Defaults to None

        Returns:
            True if the function completed successfully, False if the function timed out.
        """

        async def _dialog_coro(times):
            self.log("Going through dialog")
            while times >= 1:
                times -= 1
                while (not self.is_press_x()) and (not self.is_dialog_more()):
                    await asyncio.sleep(0.5)

                if self.is_press_x():
                    await self.send_key("X", 0.1)

                while not self.is_dialog_more():
                    await asyncio.sleep(0.5)

                while self.is_dialog_more():
                    await self.send_key("SPACEBAR", 0.1)
                    await asyncio.sleep(0.1)

            await asyncio.sleep(1)

        # run it with the timeout
        try:
            await asyncio.wait_for(_dialog_coro(times), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False

    async def logout_and_in(self, confirm=False, *, confirm_timeout=None, timeout=None):
        """
        Logs the user out and then logs it in again.

        Args:
            confirm (bool, optional): Should the bot wait for a "confirm" prompt before continuing. Defaults to False
                Set to True if logging out durring a battle or to exit an unfinished dungeon.
            confirm_timeout: timeout argument for the ``click_confirm`` task. Defaults to None
            timeout: time in seconds before the function times out. Defaults to None

        Returns:
            True if the function completed successfully, False if the function timed out.
        """

        async def _coro():
            self.log("Logging out")
            await self.send_key("ESC", 0.1)
            await self.mouse.click(259, 506, delay=0.3)
            if confirm:
                await self.click_confirm(timeout=confirm_timeout)
            # wait for player select screen
            self.log("Wait for loading")
            while not (
                self.pixel_matches_color((361, 599), (133, 36, 62), tolerance=20)
            ):
                await self.wait(0.5)

            self.log("Logging back in")
            await self.mouse.click(395, 594)
            await self.finish_loading()
            if self.is_crown_shop():
                await self.wait(0.5)
                await self.send_key("ESC", 0.1)
                await self.send_key("ESC", 0.1)

        # run it with the timeout
        try:
            await asyncio.wait_for(_coro(), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False

    async def press_x(self, *, timeout=None):
        """
        Waits for the "press x" prompt and sends the "X" key.

        Args:
            timeout (optional): value in seconds to timeout if it hasn't finished yet. Defaults to None

        Returns:
            True if the function completed successfully, False if the function timed out.
        """

        async def _press_x_coro():
            while not self.is_press_x():
                await self.wait(0.5)
            await self.send_key("X", 0.1)

        # run it with the timeout
        try:
            await asyncio.wait_for(_press_x_coro(), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False

    async def click_confirm(self, *, timeout=None):
        """
        Waits for the "confirm" prompt, and clicks "confirm"

        Args:
            timeout (optional): value in seconds to timeout if it hasn't finished yet. Defaults to None

        Returns:
            True if the function completed successfully, False if the function timed out.
        """

        async def _confirm_coro():
            await self.wait(0.2)  #
            confirm = self.get_confirm()
            while not confirm:
                await self.wait(0.5)
                confirm = self.get_confirm()

            await self.mouse.click(*confirm, duration=0.2, delay=0.2)
            await self.wait(0.5)

        # run it with the timeout
        try:
            await asyncio.wait_for(_confirm_coro(), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False

    """
    POSITION & MOVEMENT
    """

    async def get_quest_xyz(self) -> tuple:
        """
        Gets the X, Y, Z coordinates to the quest destination
        Requires the ``quest_struct`` hook to be activated

        Returns:
            tuple: (X, Y, Z) value of the quest goal location.
        """
        return await self.walker.quest_xyz()

    async def get_player_location(self) -> XYZYaw:
        """
        Fetches the player's XYZYaw location
        Requires the `player_struct` hook to be activated

        Returns:
            XYZYaw: (X, Y, Z, Yaw) tuple values of the player's position and direction.
        """
        xyz = await self.walker.xyz()
        yaw = await self.walker.yaw()
        return XYZYaw(x=xyz.x, y=xyz.y, z=xyz.z, yaw=yaw)

    async def teleport_to(self, location: XYZYaw):
        """
        Teleports to XYZYaw location
        Will return immediately if player movement is locked
        Requires the ``player_struct`` hook to be activated

        Args:
            location (XYZYaw): location to move to
        """
        if await self.walker.move_lock():
            return

        await self.walker.teleport(**location._asdict())
        await self.send_key("W", 0.1)

    async def walk_to(self, location: XYZYaw, mount_speed: float = -1):
        """
        Walks to XYZYaw location in a **straight** line only.
        Will _not_ work if there are obstacles in the way. Ideal for short distances
        Will return immediately if player movement is locked
        Requires the `player_struct` hook to be activated.


        Args:
            location (XYZYaw): The location to walk to.
                (Only the x and y value will actually be used)
            mount_speed (float): mount speed multiplier. Defaults to 1.4 (40% speed mount).
                Default can be changed by setting the ``wizsdk.client.DEFAULT_MOUNT_SPEED`` value
        """
        if mount_speed == -1:
            mount_speed = DEFAULT_MOUNT_SPEED

        if mount_speed:
            wizwalker.client.WIZARD_SPEED = 580 * mount_speed

        if await self.walker.move_lock():
            return

        await self.walker.goto(location.x, location.y)

    async def teleport_to_friend(self, match_img) -> bool:
        """
        Completes a set of actions to teleport to a friend.
        The friend must have the proper symbol next to it.
        The symbol must match the image passed as 'match_img'.

        Args:
            match_img: A string of the image file name, or a list of bytes returned by ``Client.get_image``
                The friend icon to find to select which friend to teleport to.

        Returns:
            bool: Whether the friend was found or not.
        """
        if not self.silent_mouse:
            self.set_active()
        # Check if friends already opened (and close it)
        while self.pixel_matches_color((780, 364), (230, 0, 0), 40):
            await self.mouse.click(780, 364, duration=0.1)
            await self.wait(0.2)

        # Open friend menu
        await self.mouse.click(780, 50, duration=0.2)
        await self.wait(0.2)

        # Find friend that matches friend match_img
        found = False
        last_page = False
        while (not found) and (not last_page):
            last_page = not self.pixel_matches_color((775, 328), (206, 44, 24), 50)
            found = self.locate_on_screen(match_img, region=AREA_FRIENDS, threshold=0.2)

            if (not found) and not last_page:
                await self.mouse.click(775, 328, duration=0.2)
                await self.mouse.move_to(775, 328, 0.3)

        if found is not False:
            _, y = found

            # Select friend
            await self.mouse.click(670, y, duration=0.2, delay=0.5)
            # Select port
            await self.mouse.click(450, 115, duration=0.2, delay=0.5)
            # Select yes
            await self.click_confirm()
            await self.wait(1)

            return True
        else:
            self.log("Friend could not be found")
            return False

    async def face_quest_destination(self) -> None:
        """
        Changes the player's yaw to be facing the quest destination.
        *Note:* depending on your location, this may differ from where your quest arrow is pointing to
        """
        xyz = await self.walker.xyz()
        quest_xyz = await self.walker.quest_xyz()
        yaw = calculate_perfect_yaw(xyz, quest_xyz)
        await self.walker.set_yaw(yaw)

    """
    BATTLE ACTIONS & METHODS
    """

    def get_battle(self, name: str = None) -> Battle:
        """
        Fetch a ``battle`` associated with the client

        Args:
            name (str): Name of battle for logging purposes.

        Returns:
            Battle: object with battle methods linked to this client
        """
        return Battle(self, name, default_image_folder=self._default_image_folder)

    async def find_spell(
        self, spell_name: str, threshold: float = 0.12, ignore_gray_detection=False
    ) -> Card:
        """
        Searches spell area for an image matching ``spell_name``. An additional check to see if the spell is grayed out is done by default.

        Args:
            spell_name (str): The name of the spell as you have it saved in your spells image folder.
            threshold (float): How precise the match should be. The lower this value, the more exact the match will be.
            ignore_gray_detection (bool): should the gray detection be ignored. defaults to False

        Returns:
            int: x positions of spell if found, None otherwise
        """
        # Move into the window if the window isn't active
        if not self.silent_mouse and not self.is_active():
            self.set_active()
            await self.mouse.move_to(100, 100, duration=0.2)
        else:
            # Move mouse out of area to get a clear image
            await self.mouse.move_out(AREA_SPELLS)

        # Get screenshot of `spell_area`
        b_spell_area = self.get_image(AREA_SPELLS)

        extensions = [".png", ".jpg", "jpeg", ".bmp"]
        file_name = spell_name
        if not spell_name[-4:] in extensions:
            file_name += ".png"

        spell_path = os.path.join(SPELLS_FOLDER, file_name)

        res = match_image(b_spell_area, spell_path, threshold)

        if res:
            x, y = res
            # We're only interested in the x position
            offset_x = AREA_SPELLS[0]
            # a card width is 52 pixels, round to the nearest 1/2 card (26 pixels)
            adjusted_x = round(x / 26) * 26

            spell_pos = offset_x + adjusted_x

            if not ignore_gray_detection:
                # Check if the card is grayed out
                grayness = self.is_gray_rect(
                    (spell_pos - 10, 310, 20, 20), threshold=25
                )
                # print(file_name, grayness)
                if grayness < 25:
                    if grayness > 20:
                        print(f"{file_name} was found, but gray was detected.")
                        print("If this is an error, contact wizSDK dev.")
                    return None

            return Card(self, spell_name, spell_pos)
        else:
            return None

    async def pass_turn(self) -> None:
        """
        Clicks `pass` while in a battle
        """
        await self.mouse.click(254, 398, duration=0.2, delay=0.5)
        await self.wait(0.5)

    async def autocast(self, *spells: str, target=None):
        """
        Short-hand for ``find_spell`` followed by ``cast_spell``. Finds and casts spells. If 2 spells are provided, it will enchant spell 2 with spell 1. If target is provided, it will click the target. If any of the spells are not found, this function exits with False.

        Args:
            spells: provide up to 2 spell arguments
            target (int, optional): the target to cast the spell on. See ``Card.cast``

        Returns:
            True if all spells were found, False otherwise

        Examples:
            .. code-block:: py

                player = Client.register(name="Bot")
                battle = player.get_battle("Test")
                # Loop
                while await battle.loop():
                    # The following are all correct ways of using `autocast`
                    await player.autocast("epic", "bat", target=0)
                    await player.autocast("epic", "tempest")
                    await player.autocast("storm-blade", target=4)
        """

        if len(spells) == 0:
            print(f"Invalid call to `autocast`, expected 1-3 arguments, received none")
            return

        elif type(spells[-1]) == int:
            # The 3rd argument becomes the target
            target = spells[-1]
            spells = spells[:-1]

        found = [await self.find_spell(s) for s in spells[:2]]

        # Return False if some spells weren't found
        if sum([bool(c) for c in found]) != len(found):
            return False

        if len(found) == 2:
            to_cast = await found[0].enchant(found[1])
        else:
            to_cast = found[0]

        await to_cast.cast(target=target)
        return True


def register_clients(
    n_handles_expected: int,
    names: list = [],
    confirm_position: bool = False,
    silent_mouse: bool = False,
) -> list:
    """
    Register multiple clients, sorted from left to right, top to bottom.

    Args:
        n_handles_expected (int): the expected # of wiz windows opened. Use -1 for undetermined
        names (list): A list of strings that will serve as the names of the windows
        confirm_position (bool): prompt the user to confirm the windows order before continuing
        silent_mouse: When enabled, moves the mouse without taking control of the actual cursor

    Returns:
        client_list (list): A list populated with ``Client`` instances
    """
    accepted = False
    while not accepted:
        handles = get_all_wiz_handles()
        n_handles = len(handles)

        if n_handles != n_handles_expected and n_handles_expected > 0:
            print(
                f"Invalid number of windows open. {n_handles_expected} required, {n_handles} detected."
            )
            os.system("pause")
            exit()
        else:
            print(f"{n_handles} windows detected")

        # Fill names array if necessary
        for i in range(n_handles - len(names)):
            names.append(None)

        # Register and order the windows from left to right, top to bottom
        w = [
            Client.register(handle=handles[i], silent_mouse=silent_mouse)
            for i in range(n_handles)
        ]

        # Sort
        def sort_func(win):
            rect = win.get_rect()
            round_y = (rect[1] // 100) * 100
            return rect[0] + (round_y * 10)

        w.sort(key=sort_func)

        # Set names
        for i in range(len(w)):
            if i < len(names):
                w[i].set_name(names[i])

        # Confirm position
        if confirm_position:
            print("Is this order ok?")
            answer = input("[y] or n: ")
            if answer.lower().strip()[0] == "y" or answer == "":
                accepted = True
            else:
                print("Re-order the windows")
                os.system("pause")
        else:
            accepted = True

    # Returns the sorted clients in an array
    return w



# Native imports
import ctypes
import ctypes.wintypes
from os import path

# Third-party imports
import numpy as np
import cv2

# Custom imports
from .window import Window


user32 = ctypes.WinDLL("user32.dll")
gdi32 = ctypes.WinDLL("gdi32.dll")

# Converted to python from https://docs.microsoft.com/en-us/windows/win32/gdi/capturing-an-image
# by Starrfox


class _BITMAPINFOHEADER(ctypes.Structure):
    _fields_ = [
        ("biSize", ctypes.c_uint32),
        ("biWidth", ctypes.c_int),
        ("biHeight", ctypes.c_int),
        ("biPlanes", ctypes.c_short),
        ("biBitCount", ctypes.c_short),
        ("biCompression", ctypes.c_uint32),
        ("biSizeImage", ctypes.c_uint32),
        ("biXPelsPerMeter", ctypes.c_long),
        ("biYPelsPerMeter", ctypes.c_long),
        ("biClrUsed", ctypes.c_uint32),
        ("biClrImportant", ctypes.c_uint32),
    ]


least_gray = 0


class _BITMAP(ctypes.Structure):
    _fields_ = [
        ("bmType", ctypes.c_long),
        ("bmWidth", ctypes.c_long),
        ("bmHeight", ctypes.c_long),
        ("bmWidthBytes", ctypes.c_long),
        ("bmPlanes", ctypes.wintypes.WORD),
        ("bmBitsPixel", ctypes.wintypes.WORD),
        ("bmBits", ctypes.wintypes.LPVOID),
    ]


class DeviceContext(Window):
    """
    Base class for accessing the Window's Device Context (pixels, image captures, etc..)
    """

    def __init__(self, handle):
        super().__init__(handle)
        self.window_handle = handle

    def get_image(self, region=None):
        """
        returns a byte array with the pixel data of the ``region`` from the ``window_handle`` window. ``region`` is relative to the ``window_handle`` window. If no ``region`` is specified, it will capture the entire window. If no ``window_handle`` is provided on initiation, monitor 1 is used as the context.

        Args:
            region: (x, y, width, height) tuple relative to the ``window_handle`` context. Defaults to None

        Returns:
            A 2d numpy array representing the pixel data of the captured region.
        """

        _, _, w, h = self.get_rect()
        x, y = 0, 0

        if region and len(region) == 4:
            x, y, w, h = region

        # Get devices context
        wDC = user32.GetWindowDC(self.window_handle)
        if wDC == 0:
            print("Window handle retrieval error")
            return []

        # Where we will move the pixels to
        mDC = gdi32.CreateCompatibleDC(wDC)
        if mDC == 0:
            print("Memory device creation error")
            return []

        gdi32.SetStretchBltMode(wDC, 4)
        # Create empty bitmap
        mBM = gdi32.CreateCompatibleBitmap(wDC, w, h)
        if mBM == 0:
            print("Bitmap creation error")
            return []

        # Select wDC into bitmap
        gdi32.SelectObject(mDC, mBM)

        gdi32.BitBlt(mDC, 0, 0, w, h, wDC, x, y, 0x00CC0020)

        bitmap = _BITMAP()

        bits_transfered = 0
        while bits_transfered == 0:
            bits_transfered = gdi32.GetObjectA(
                mBM, ctypes.sizeof(_BITMAP), ctypes.byref(bitmap)
            )

        bi = _BITMAPINFOHEADER()
        bi.biSize = ctypes.sizeof(_BITMAPINFOHEADER)
        bi.biWidth = bitmap.bmWidth
        bi.biHeight = bitmap.bmHeight
        bi.biPlanes = 1
        bi.biBitCount = 32
        bi.biCompression = 0
        bi.biSizeImage = 0
        bi.biXPelsPerMeter = 0
        bi.biYPelsPerMeter = 0
        bi.biClrUsed = 0
        bi.biClrImportant = 0

        bitmap_size = (
            (
                (bitmap.bmWidth * bitmap.bmBitsPixel + bitmap.bmBitsPixel - 1)
                // bitmap.bmBitsPixel
            )
            * 4
            * bitmap.bmHeight
        )

        bitmap_buffer = (ctypes.c_char * bitmap_size)()

        gdi32.GetDIBits(
            wDC,
            mBM,
            0,
            bitmap.bmHeight,
            ctypes.byref(bitmap_buffer),
            ctypes.byref(bi),
            0,
        )

        gdi32.DeleteObject(mBM)
        gdi32.DeleteDC(mDC)
        user32.ReleaseDC(self.window_handle, wDC)

        img = np.frombuffer(bitmap_buffer.raw, dtype="uint8")
        img = img.reshape((bitmap.bmHeight, bitmap.bmWidth, 4))
        img = np.flip(img, 0)
        # Remove the alpha channel
        img = img[:, :, :3]
        return img

    def get_pixel(self, x, y) -> tuple:
        """
        Returns the (red, green, blue) channel's of the pixel at ``x``, ``y`` relative to the ``window_handle`` context.

        Args:
            x
            y

        Returns:
            (r, g, b) tuple
        """
        hDC = user32.GetWindowDC(self.window_handle)
        rgb = gdi32.GetPixel(hDC, x, y)
        user32.ReleaseDC(self.window_handle, hDC)
        r = rgb & 0xFF
        g = (rgb >> 8) & 0xFF
        b = (rgb >> 16) & 0xFF
        return (r, g, b)

    def screenshot(self, filename, region=None):
        """
        captures a screenshot of the provided ``region``, saves it to file as ``filename``.

        Args:
            filename: what to save to image as
            region: (x, y, width, height) tuple relative to the ``window_handle`` context. Defaults to None
        """
        image = self.get_image(region=region)
        cv2.imshow("mat", image)
        cv2.waitKey(0)
        cv2.imwrite(filename, image)

    def pixel_matches_color(self, xy, expected_rgb, tolerance=0):
        """
        gets the value of a pixel with ``get_pixel`` and checks it against ``expected_rgb``. Accepts ``tolerance`` amount of differences between the pixel and its expected value.
        """
        pixel = self.get_pixel(*xy)
        if len(pixel) == 3 or len(expected_rgb) == 3:  # RGB mode
            r, g, b = pixel[:3]
            exR, exG, exB = expected_rgb[:3]
            return (
                (abs(r - exR) <= tolerance)
                and (abs(g - exG) <= tolerance)
                and (abs(b - exB) <= tolerance)
            )
        else:
            assert (
                False
            ), f"Color mode was expected to be length 3 (RGB), but pixel is length {len(pix)} and expected_RGB is length { len(expected_rgb)}"

    def is_gray_rect(self, region, threshold=25):
        """
        calculates if a ``(x, y, width, height)`` ``region`` is gray by iterating through all its pixels and calculating the difference between the channel with the lowest value, and the one with the highest value. Stops iterating if that value is greater than ``threshold``. Returns the highest of the values calculated.

        Args:
            region: (x, y, width, height) tuple relative to the ``window_handle`` context.
            threshold: difference allowed between highest channel and lowest channel to still be considered gray.

        Returns:
            the greatest difference between the highest channel and lowest channel.
        """
        # global least_gray
        least_gray = 0

        w, h = region[2:]
        img = self.get_image(region)

        gray = True

        # Check if all pixels in image are gray
        for x in range(h):
            if not gray:
                break

            for y in range(w):
                pixel = img[x][y]

                # Determine if a pixel is gray enough
                color = abs(int(min(*pixel)) - int(max(*pixel)))
                if color > least_gray:
                    least_gray = color

                if color > threshold:
                    gray = False
                    break

        return least_gray

    def locate_on_screen(
        self, match_img, region=None, *, threshold=0.1, debug=False, folder=None
    ):
        """
        Attempts to locate `match_img` in the Wizard101 window.
        pass a rect tuple `(x, y, width, height)` as the `region` argument to narrow
        down the area to look for the image.
        Adjust `threshold` for the precision of the match (between 0 and 1, the lowest being more precise).
        Set `debug` to True for extra debug info

        Args:
            match_img: to image to locate, can be a file name or a numpy array
            region: (x, y, width, height) tuple relative to the ``window_handle`` context. Defaults to None
            theshold: precision of the match -- between 0 and 1, the lowest being more precise
            debug: set to True to show a pop up of the area that matched the image provided.
            folder: folder to look in. Overrides ``IMAGE_FOLDER`` default

        Returns:
            (x, y) tuple for center of match if found. False otherwise.

        """
        to_match = (
            path.join(folder or self._default_image_folder or "", match_img)
            if type(match_img) == str
            else match_img
        )
        match = match_image(
            self.get_image(region=region), to_match, threshold, debug=debug
        )

        if not match or not region:
            return match

        region_x, region_y = region[:2]
        x, y = match
        return x + region_x, y + region_y


def _to_cv2_img(data):
    if type(data) is str:
        # It's a file name
        # cv2.IMREAD_COLOR ignores alpha channel, loads only rgb
        img = cv2.imdecode(np.fromfile(data, dtype=np.uint8), cv2.IMREAD_COLOR)

        return img

    elif type(data) is np.ndarray:
        # It's a np array
        return data

    return None


def match_image(largeImg, smallImg, threshold=0.1, debug=False):
    """
    Finds smallImg in largeImg using template matching
    Adjust threshold for the precision of the match (between 0 and 1, the lowest being more precise)

    Returns:
        tuple (x, y) of the center of the match if it's found, False otherwise.
    """

    method = cv2.TM_SQDIFF_NORMED

    small_image = _to_cv2_img(smallImg)
    large_image = _to_cv2_img(largeImg)

    if (small_image is None) or (large_image is None):
        print("Error: large_image or small_image is None")
        return False

    h, w = small_image.shape[:-1]

    if debug:
        print("large_image:", large_image.shape)
        print("small_image:", small_image.shape)

    try:
        result = cv2.matchTemplate(small_image, large_image, method)
    except cv2.error as e:
        # The image was not found. like, not even close. :P
        print(e)
        return False

    # We want the minimum squared difference
    mn, _, mnLoc, _ = cv2.minMaxLoc(result)

    if mn >= threshold:
        if debug:
            cv2.imshow("output", large_image)
            cv2.waitKey(0)
        return False

    # Extract the coordinates of our best match
    x, y = mnLoc

    if debug:
        print(f"Match at ({x}, {y}) relative to region")
        # Draw the rectangle:
        # Get the size of the template. This is the same size as the match.
        trows, tcols = small_image.shape[:2]

        # If I don't call this a get a TypeError :P
        large_image = np.array(large_image)
        # Draw the rectangle on large_image
        cv2.rectangle(large_image, (x, y), (x + tcols, y + trows), (0, 0, 255), 2)

        # Display the original image with the rectangle around the match.
        cv2.imshow("output", large_image)

        # The image is only displayed if we call this
        cv2.waitKey(0)

    # Return coordinates to center of match
    return (x + (w // 2), y + (h // 2))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wizsdk']

package_data = \
{'': ['*'], 'wizsdk': ['images/*']}

install_requires = \
['Sphinx>=3.4.3,<4.0.0',
 'asyncchain>=0.1.0,<0.2.0',
 'asyncio>=3.4.3,<4.0.0',
 'numpy>=1.19.1,<2.0.0',
 'opencv-python>=4.4.0,<5.0.0',
 'wizwalker>=0.16.3,<0.17.0']

setup_kwargs = {
    'name': 'wizsdk',
    'version': '1.3.8',
    'description': 'API for interacting with and making bots for wizard101',
    'long_description': "# WizSDK\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nWizard101 bot Software Development Kit\nA wrapper for [WizWalker](https://github.com/StarrFox/wizwalker) bundled with many other useful function.\nDesigned to make building bots for farming / questing / bazaar sniping / selling / pet training\n\n## documentation\n\nyou can find the documentation [here](https://underpaiddev1.github.io/wizSDK/)\n\nyou can download these from the gh-pages branch if desired\n\n## install\n\nclone the repo or install from pypi `pip install -U wizsdk`\n\n## discord\n\njoin the offical discord [here](https://discord.gg/D9GRrbDzpt)\n\n## development install\n\nThis package uses [poetry](https://python-poetry.org/)\n\n```shell script\n$ poetry install\n```\n\n## building\n\nYou'll need the dev install (see above) for this to work\n\n### Docs\n\n```shell script\n$ cd docs\n$ make html\n```\n",
    'author': 'Underpaid Dev',
    'author_email': 'underpaiddev1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
